/*
 * The contents of this file are subject to the Mozilla Public License Version 1.1
 * (the "License"); you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at <http://www.mozilla.org/MPL/>.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, WITHOUT
 * WARRANTY OF ANY KIND, either express or implied. See the License for the specific
 * language governing rights and limitations under the License.
 * 
 * The Original Code is the Venice Web Communities System.
 * 
 * The Initial Developer of the Original Code is Eric J. Bowersox <erbo@silcom.com>,
 * for Silverwrist Design Studios.  Portions created by Eric J. Bowersox are
 * Copyright (C) 2001 Eric J. Bowersox/Silverwrist Design Studios.  All Rights Reserved.
 * 
 * Contributor(s): 
 */
package com.silverwrist.venice.servlets.format;

import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import org.apache.log4j.*;
import org.w3c.dom.*;
import com.silverwrist.util.*;
import com.silverwrist.venice.core.CommunityContext;
import com.silverwrist.venice.core.UserContext;
import com.silverwrist.venice.core.VeniceEngine;
import com.silverwrist.venice.except.*;
import com.silverwrist.venice.servlets.Variables;
import com.silverwrist.venice.servlets.format.menus.*;
import com.silverwrist.venice.servlets.format.sideboxes.SideBoxFactory;
import com.silverwrist.venice.util.XMLLoader;

public class RenderConfig implements ColorSelectors
{
  /*--------------------------------------------------------------------------------
   * Static data values
   *--------------------------------------------------------------------------------
   */

  protected static final String ATTR_NAME = "com.silverwrist.venice.servlets.RenderConfig";
  protected static final String CONFIG_FILE_PARAM = "render.config";

  private static Category logger = Category.getInstance(RenderConfig.class);

  private static Map colornames_map;

  /*--------------------------------------------------------------------------------
   * Attributes
   *--------------------------------------------------------------------------------
   */

  private String site_title;
  private boolean want_comments;
  private boolean allow_gzip;
  private boolean no_smart_tags;
  private String font_face;
  private File stylesheet;
  private long stylesheet_time = 0;
  private String image_url;
  private String static_url;
  private String site_logo;
  private int site_logo_width = 140;
  private int site_logo_height = 80;
  private String site_logo_linkURL = null;
  private StockMessages stock_messages;
  private Map menus;
  private String[] colors_array;
  private int footer_logo_scale;
  private Map sidebox_factories;
  private String photo_not_avail;
  private boolean photo_not_avail_fixup;
  private CommunityLeftMenuFactory comm_menu_fact;

  /*--------------------------------------------------------------------------------
   * Constructor
   *--------------------------------------------------------------------------------
   */

  protected RenderConfig(String config_file, String root_file_path) throws ConfigException
  {
    XMLLoader loader = XMLLoader.get();
    Document doc = loader.loadConfigDocument(config_file);

    // Make sure the document is valid.
    Element root = loader.configGetRootElement(doc,"render-config");
    DOMElementHelper root_h = new DOMElementHelper(root);

    // Get the site title.
    site_title = loader.configGetSubElementText(root_h,"site-name");
    if (logger.isDebugEnabled())
      logger.debug("Site title: " + site_title);

    // Get the <rendering/> section.
    Element sect = loader.configGetSubSection(root_h,"rendering");

    // Load the quick Boolean values indicated by the presence or absence of a tag.
    DOMElementHelper sect_h = new DOMElementHelper(sect);
    want_comments = sect_h.hasChildElement("html-comments");
    allow_gzip = sect_h.hasChildElement("gzip-output");
    no_smart_tags = !(sect_h.hasChildElement("ms-copyright-violations"));
    if (logger.isDebugEnabled())
    { // log the read values
      logger.debug("Use HTML comments: " + want_comments);
      logger.debug("Use GZIP encoding: " + allow_gzip);
      logger.debug("Disable IE Smart Tags: " + no_smart_tags);

    } // end if

    // Load the default font face.
    font_face = loader.configGetSubElementText(sect_h,"font");
    if (logger.isDebugEnabled())
      logger.debug("Font face: " + font_face);

    // Load the default stylesheet file reference.
    String tmp = sect_h.getSubElementText("stylesheet");
    if (tmp!=null)
    { // we're using Cascading Stylesheets - load it and test for existence
      if (!(tmp.startsWith("/")))  // prepend app root
	tmp = root_file_path + tmp;
      if (logger.isDebugEnabled())
	logger.debug("Stylesheet location: " + tmp);

      // Test to make sure the stylesheet is actually present.
      stylesheet = new File(tmp);
      if (!(stylesheet.exists() && stylesheet.canRead()))
      { // it's not there - bail out!
	logger.fatal("unable to read stylesheet file: " + tmp);
	throw new ConfigException("stylesheet " + tmp + " cannot be read",sect);

      } // end if

    } // end if
    else // no stylesheet
      stylesheet = null;

    // Get the <colors/> section and load the colors out of it..
    Element sect1 = loader.configGetSubSection(sect_h,"colors");
    colors_array = new String[colornames_map.size()];
    int i;
    NodeList nl = sect1.getChildNodes();
    for (i=0; i<nl.getLength(); i++)
    { // look at all subelements and map their names to color selectors
      Node n = nl.item(i);
      if (n.getNodeType()==Node.ELEMENT_NODE)
      { // map the name of the node to a color selector
	Integer cs = (Integer)(colornames_map.get(n.getNodeName()));
	if (cs!=null)
	{ // get the color value and load it into the array
	  DOMElementHelper h = new DOMElementHelper((Element)n);
	  colors_array[cs.intValue()] = h.getElementText();

	} // end if

      } // end if

    } // end for

    // Load in the footer logo scale.
    Integer itmp = sect_h.getSubElementInt("footer-logo-scale");
    footer_logo_scale = ((itmp==null) ? 100 : itmp.intValue());

    // Get the <paths/> section.
    sect = loader.configGetSubSection(root_h,"paths");
    sect_h = new DOMElementHelper(sect);

    // Load the image path.
    image_url = loader.configGetSubElementText(sect_h,"image");
    if (logger.isDebugEnabled())
      logger.debug("Image path: " + image_url);

    // Load the static path.
    static_url = loader.configGetSubElementText(sect_h,"static");
    if (logger.isDebugEnabled())
      logger.debug("Static files path: " + static_url);

    // Get the site logo element.
    sect1 = loader.configGetSubSection(sect_h,"site-logo");
    DOMElementHelper sect1_h = new DOMElementHelper(sect1);
    site_logo = loader.configGetText(sect1_h);

    // Get the width and height of the logo.
    itmp = sect1_h.getAttributeInt("width");
    if (itmp!=null)
      site_logo_width = itmp.intValue();
    itmp = sect1_h.getAttributeInt("height");
    if (itmp!=null)
      site_logo_height = itmp.intValue();

    // Get the link URL of the logo.
    tmp = sect1.getAttribute("href");
    if (!(StringUtil.isStringEmpty(tmp)))
      site_logo_linkURL = tmp;

    if (logger.isDebugEnabled())
      logger.debug("Site logo: " + site_logo);

    // Get the name of the sidebox configuration file.
    String sidebox_config = loader.configGetSubElementText(sect_h,"sidebox-config");
    if (!(sidebox_config.startsWith("/")))
      sidebox_config = root_file_path + sidebox_config;

    // Get the name of the "photo not available" image.
    sect1 = sect_h.getSubElement("photo-not-avail");
    if (sect1!=null)
    { // load the element text and fixup information
      sect1_h = new DOMElementHelper(sect1);
      photo_not_avail = sect1_h.getElementText();
      photo_not_avail_fixup = sect1_h.hasAttribute("fixup");

    } // end if
    else
    { // just load the defaults
      photo_not_avail = "photo_not_avail.gif";
      photo_not_avail_fixup = true;

    } // end else

    // Get the name of the services configuration file.
    String services_config = loader.configGetSubElementText(sect_h,"services-config");
    if (!(services_config.startsWith("/")))
      services_config = root_file_path + services_config;

    // Load the <messages/> section.
    sect = loader.configGetSubSection(root_h,"messages");

    // Initialize the stock messages list.
    stock_messages = new StockMessages(sect);

    // Load the <menu-definitions/> section.
    sect = loader.configGetSubSection(root_h,"menu-definitions");

    // Initialize the menus list.
    HashMap tmp_menus = new HashMap();
    nl = sect.getChildNodes();
    for (i=0; i<nl.getLength(); i++)
    { // look for <menudef> subnodes and use them to initialize menus
      Node n = nl.item(i);
      if (n.getNodeType()==Node.ELEMENT_NODE)
      { // verify that it's a menu definition, then get its ID and build a menu
	loader.configVerifyNodeName(n,"menudef",sect);
	String menuid = loader.configGetAttribute((Element)n,"id");

	// create the menu and add it to the mapping
	LeftMenu menu = new LeftMenu((Element)n,menuid);
	tmp_menus.put(menuid,menu);
	if (logger.isDebugEnabled())
	  logger.debug("menu \"" + menuid + "\" defined");

      } // end if
      // else just ignore it
      
    } // end for

    // save off the menus as an unmodifiable map
    if (tmp_menus.isEmpty())
      menus = Collections.EMPTY_MAP;
    else
      menus = Collections.unmodifiableMap(tmp_menus);
    if (logger.isDebugEnabled())
      logger.debug(menus.size() + " menu definitions loaded from config");

    // done with the render-config.xml file
    // Load up the sidebox-config.xml file.
    doc = loader.loadConfigDocument(sidebox_config);
    root = loader.configGetRootElement(doc,"sidebox-config");

    // Examine the child nodes for sidebox configuration data.
    nl = root.getChildNodes();
    HashMap tmp_factories = new HashMap();
    for (i=0; i<nl.getLength(); i++)
    { // extract each sidebox section from the list and build a new master sidebox object
      Node n = nl.item(i);
      if ((n.getNodeType()==Node.ELEMENT_NODE) && (n.getNodeName().equals("sidebox")))
      { // get the ID for this sidebox
	try
	{ // get the ID and convert it to integer
	  tmp = loader.configGetAttribute((Element)n,"id");
	  itmp = new Integer(tmp);

	} // end try
	catch (NumberFormatException nfe)
	{ // sidebox ID was not an integer!
	  logger.fatal("<sidebox/> ID not numeric!");
	  throw new ConfigException("non-numeric ID specified in <sidebox/>",(Element)n);

	} // end catch

	SideBoxFactory factory = createSideBoxFactory((Element)n);
	tmp_factories.put(itmp,factory);

      } // end if
      // else ignore this node

    } // end for

    if (tmp_factories.isEmpty())
      sidebox_factories = Collections.EMPTY_MAP;
    else
      sidebox_factories = Collections.unmodifiableMap(tmp_factories);

    // done with the sidebox-config.xml file
    // Load up the services-config.xml file.
    doc = loader.loadConfigDocument(services_config);
    root = loader.configGetRootElement(doc,"services-config");
    root_h = new DOMElementHelper(root);

    // Get the community section and pass it to the CommunityLeftMenuFactory.
    comm_menu_fact = new CommunityLeftMenuFactory(root_h.getSubElement("community"));

  } // end constructor

  /*--------------------------------------------------------------------------------
   * Internal functions
   *--------------------------------------------------------------------------------
   */

  private static SideBoxFactory createSideBoxFactory(Element cfg) throws ConfigException
  {
    XMLLoader loader = XMLLoader.get();
    String cname = loader.configGetSubElementText(cfg,"factory-class");

    try
    { // load the factory class and create an instance of the factory
      SideBoxFactory factory = (SideBoxFactory)(Class.forName(cname).newInstance());
      factory.setConfiguration(cfg);
      return factory;

    } // end try
    catch (ClassNotFoundException cnfe)
    { // sidebox could not be configured
      logger.fatal("<sidebox/> config <factory-class/> not valid!",cnfe);
      throw new ConfigException("<factory-class/> in <sidebox/> is not a valid class!",cfg);

    } // end catch
    catch (IllegalAccessException iae)
    { // could not access class and/or constructor
      logger.fatal("<sidebox/> <factory-class/> is not accessible!",iae);
      throw new ConfigException("<factory-class/> in <sidebox/> is not accessible!",cfg);

    } // end catch
    catch (InstantiationException ie)
    { // unable to create the class
      logger.fatal("<sidebox/> <factory-class/> could not be instantiated!",ie);
      throw new ConfigException("<factory-class/> in <sidebox/> could not be created!",cfg);

    } // end catch
    catch (ClassCastException cce)
    { // the class is not a SideBoxFactory implementor - cannot create
      logger.fatal("<sidebox/> <factory-class/> is not a SideBoxFactory!",cce);
      throw new ConfigException("<factory-class/> in <sidebox/> is not the correct type!",cfg);

    } // end catch

  } // end createSideBoxFactory

  /*--------------------------------------------------------------------------------
   * External operations usable only by RenderData
   *--------------------------------------------------------------------------------
   */

  boolean useHTMLComments()
  {
    return want_comments;

  } // end useHTMLComments

  boolean isGZIPAllowed()
  {
    return allow_gzip;

  } // end isGZIPAllowed

  boolean noSmartTags()
  {
    return no_smart_tags;

  } // end noSmartTags

  String getFullImagePath(String name)
  {
    StringBuffer buf = new StringBuffer();
    buf.append(image_url).append(name);
    return buf.toString();

  } // end getFullImagePath

  String getStaticFilePath(String name)
  {
    StringBuffer buf = new StringBuffer();
    buf.append(static_url).append(name);
    return buf.toString();

  } // end getStaticFilePath

  String getTitleTag(String specific)
  {
    StringBuffer buf = new StringBuffer();
    buf.append("<TITLE>").append(specific).append(" - ").append(site_title).append("</TITLE>");
    return buf.toString();

  } // end getTitleTag

  String getSiteImageTag(int hspace, int vspace)
  {
    StringBuffer buf = new StringBuffer();
    if (site_logo_linkURL!=null)
      buf.append("<A HREF=\"").append(site_logo_linkURL).append("\">");
    buf.append("<IMG SRC=\"").append(site_logo).append("\" ALT=\"").append(site_title);
    buf.append("\" WIDTH=").append(site_logo_width).append(" HEIGHT=").append(site_logo_height);
    buf.append(" BORDER=0");
    if (hspace>0)
      buf.append(" HSPACE=").append(hspace);
    if (vspace>0)
      buf.append(" VSPACE=").append(vspace);
    buf.append('>');
    if (site_logo_linkURL!=null)
      buf.append("</A>");
    return buf.toString();

  } // end getSiteImageTag

  String getStdFontTag(String color, int size)
  {
    StringBuffer buf = new StringBuffer("<FONT FACE=\"");
    buf.append(font_face).append("\" SIZE=").append(size);
    if (color!=null)
      buf.append(" COLOR=\"").append(color).append("\"");
    buf.append('>');
    return buf.toString();

  } // end getStdFontTag

  String getStdFontTag(int selector, int size)
  {
    if ((selector<0) || (selector>=colors_array.length))
      throw new IndexOutOfBoundsException("getStdColor(): invalid color selector value");
    return getStdFontTag(colors_array[selector],size);

  } // end getStdFontTag

  String getStdBaseFontTag(int size)
  {
    StringBuffer buf = new StringBuffer("<BASEFONT FACE=\"");
    buf.append(font_face).append("\" SIZE=").append(size).append('>');
    return buf.toString();

  } // end getStdBaseFontTag

  public String getRequiredBullet()
  {
    StringBuffer buf = new StringBuffer("<FONT FACE=\"");
    buf.append(font_face).append("\" COLOR=\"red\">*</FONT>");
    return buf.toString();

  } // end getRequiredBullet

  void writeContentHeader(Writer out, String primary, String secondary) throws IOException
  {
    out.write("<SPAN CLASS=\"chead1\">" + getStdFontTag(colors_array[CONTENT_HEADER],5) + "<B>"
	      + StringUtil.encodeHTML(primary) + "</B></FONT></SPAN>");
    if (secondary!=null)
      out.write("&nbsp;&nbsp;<SPAN CLASS=\"chead2\">" + getStdFontTag(colors_array[CONTENT_HEADER],3) + "<B>"
		+ StringUtil.encodeHTML(secondary) + "</B></FONT></SPAN>");
    out.write("<HR ALIGN=LEFT SIZE=2 WIDTH=\"90%\" NOSHADE>\n");

  } // end writeContentHeader

  String getStockMessage(String identifier)
  {
    return (String)(stock_messages.get(identifier));

  } // end getStockMessage

  void writeStockMessage(Writer out, String identifier) throws IOException
  {
    String text = (String)(stock_messages.get(identifier));
    out.write(StringUtil.encodeHTML(text));

  } // end writeStockMessage

  String getStdColor(int selector)
  {
    if ((selector<0) || (selector>=colors_array.length))
      throw new IndexOutOfBoundsException("getStdColor(): invalid color selector value");
    return colors_array[selector];

  } // end getStdColor

  int scaleFooterLogo(int param)
  {
    return (param * footer_logo_scale) / 100;

  } // end scaleFooterLogo

  public LeftMenu getLeftMenu(String identifier)
  {
    return (LeftMenu)(menus.get(identifier));

  } // end getLeftMenu

  synchronized String loadStyleSheetData() throws IOException
  {
    if (stylesheet==null)
      return null;

    // Load the stylesheet data.
    StringBuffer raw_data = IOUtil.loadText(stylesheet);
    stylesheet_time = stylesheet.lastModified();
    
    // Set up the replacements map to replace the various parameters.
    HashMap vars = new HashMap();
    vars.put("font",font_face);
    vars.put("color.frame",colors_array[FRAME_BACKGROUND]);
    vars.put("color.top.background",colors_array[TITLE_BACKGROUND]);
    vars.put("color.top.foreground",colors_array[TITLE_FOREGROUND]);
    vars.put("color.top.link",colors_array[TITLE_LINK]);
    vars.put("color.left.background",colors_array[LEFT_BACKGROUND]);
    vars.put("color.left.foreground",colors_array[LEFT_FOREGROUND]);
    vars.put("color.left.link",colors_array[LEFT_LINK]);
    vars.put("color.content.background",colors_array[CONTENT_BACKGROUND]);
    vars.put("color.content.foreground",colors_array[CONTENT_FOREGROUND]);
    vars.put("color.content.header",colors_array[CONTENT_HEADER]);
    vars.put("color.disabled",colors_array[CONTENT_DISABLED]);
    vars.put("color.error",colors_array[CONTENT_ERROR]);
    vars.put("color.sidebox.top.background",colors_array[SIDEBOX_TITLE_BACKGROUND]);
    vars.put("color.sidebox.top.foreground",colors_array[SIDEBOX_TITLE_FOREGROUND]);
    vars.put("color.sidebox.background",colors_array[SIDEBOX_CONTENT_BACKGROUND]);
    vars.put("color.sidebox.foreground",colors_array[SIDEBOX_CONTENT_FOREGROUND]);
    vars.put("color.sidebox.link",colors_array[SIDEBOX_CONTENT_LINK]);
    vars.put("color.dlg.confirm.title.background",colors_array[CONFIRM_TITLE_BACKGROUND]);
    vars.put("color.dlg.confirm.title.foreground",colors_array[CONFIRM_TITLE_FOREGROUND]);
    vars.put("color.dlg.error.title.background",colors_array[ERROR_TITLE_BACKGROUND]);
    vars.put("color.dlg.error.title.foreground",colors_array[ERROR_TITLE_FOREGROUND]);
    return StringUtil.replaceAllVariables(raw_data.toString(),vars);

  } // end loadStyleSheet

  boolean hasStyleSheetChanged()
  {
    if (stylesheet==null)
      return false;
    return (stylesheet_time!=stylesheet.lastModified());

  } // end hasStyleSheetChanged

  boolean useStyleSheet()
  {
    return (stylesheet!=null);

  } // end useStyleSheet

  VeniceContent createSideBox(int id, VeniceEngine engine, UserContext uc) throws AccessError, DataException
  {
    SideBoxFactory fact = (SideBoxFactory)(sidebox_factories.get(new Integer(id)));
    if (fact==null)
      throw new DataException("invalid sidebox ID!");
    return fact.create(engine,uc);

  } // end createSideBox

  String getPhotoNotAvail()
  {
    return photo_not_avail;

  } // end getPhotoNotAvail

  boolean getPhotoNotAvailFixup()
  {
    return photo_not_avail_fixup;

  } // end getPhotoNotAvailFixup

  public CommunityLeftMenu createCommunityMenu(CommunityContext comm)
  {
    return comm_menu_fact.createMenu(comm);

  } // end createCommunityMenu

  /*--------------------------------------------------------------------------------
   * Static operations for use by VeniceServlet
   *--------------------------------------------------------------------------------
   */

  public static RenderConfig getRenderConfig(ServletContext ctxt) throws ServletException
  {
    // Look in the servlet attributes first.
    Object obj = ctxt.getAttribute(ATTR_NAME);
    if (obj!=null)
      return (RenderConfig)obj;

    // Get the root file path.
    String root_file_path = ctxt.getRealPath("/");
    if (!(root_file_path.endsWith("/")))
      root_file_path += "/";

    // Get the parameter for the renderer's config file.
    String cfgfile = ctxt.getInitParameter(CONFIG_FILE_PARAM);
    if (!(cfgfile.startsWith("/")))
      cfgfile = root_file_path + cfgfile;
    logger.info("Initializing Venice rendering using config file: " + cfgfile);
    
    try
    { // create the RenderConfig object and save it to attributes.
      RenderConfig rconf = new RenderConfig(cfgfile,root_file_path);
      ctxt.setAttribute(ATTR_NAME,rconf);
      return rconf;

    } // end try
    catch (ConfigException e)
    { // configuration failed!  post an error message
      logger.fatal("Rendering configuration failed: " + e.getMessage(),e);
      throw new ServletException("Venice rendering configuration failed: " + e.getMessage(),e);

    } // end catch

  } // end getRenderConfig

  public static RenderData createRenderData(ServletContext ctxt, HttpServletRequest request,
					    HttpServletResponse response) throws ServletException
  {
    UserContext uc = Variables.getUserContext(ctxt,request,request.getSession(true));
    return new RenderData(getRenderConfig(ctxt),uc,ctxt,request,response);

  } // end createRenderData

  public static RenderData createRenderData(ServletContext ctxt, UserContext uc, HttpServletRequest request,
					    HttpServletResponse response) throws ServletException
  {
    return new RenderData(getRenderConfig(ctxt),uc,ctxt,request,response);

  } // end createRenderData

  /*--------------------------------------------------------------------------------
   * Static initializer
   *--------------------------------------------------------------------------------
   */

  static
  { // Initialize the mapping of color names from render-config.xml to color selector IDs.
    HashMap m = new HashMap();
    m.put("frame-bg",new Integer(FRAME_BACKGROUND));
    m.put("title-bg",new Integer(TITLE_BACKGROUND));
    m.put("title-fg",new Integer(TITLE_FOREGROUND));
    m.put("title-link",new Integer(TITLE_LINK));
    m.put("left-bg",new Integer(LEFT_BACKGROUND));
    m.put("left-fg",new Integer(LEFT_FOREGROUND));
    m.put("left-link",new Integer(LEFT_LINK));
    m.put("content-bg",new Integer(CONTENT_BACKGROUND));
    m.put("content-fg",new Integer(CONTENT_FOREGROUND));
    m.put("content-hdr",new Integer(CONTENT_HEADER));
    m.put("content-disabled",new Integer(CONTENT_DISABLED));
    m.put("content-error",new Integer(CONTENT_ERROR));
    m.put("sidebox-title-bg",new Integer(SIDEBOX_TITLE_BACKGROUND));
    m.put("sidebox-title-fg",new Integer(SIDEBOX_TITLE_FOREGROUND));
    m.put("sidebox-content-bg",new Integer(SIDEBOX_CONTENT_BACKGROUND));
    m.put("sidebox-content-fg",new Integer(SIDEBOX_CONTENT_FOREGROUND));
    m.put("sidebox-content-link",new Integer(SIDEBOX_CONTENT_LINK));
    m.put("confirm-title-bg",new Integer(CONFIRM_TITLE_BACKGROUND));
    m.put("confirm-title-fg",new Integer(CONFIRM_TITLE_FOREGROUND));
    m.put("error-title-bg",new Integer(ERROR_TITLE_BACKGROUND));
    m.put("error-title-fg",new Integer(ERROR_TITLE_FOREGROUND));
    colornames_map = Collections.unmodifiableMap(m);

  } // end static initializer

} // end class RenderConfig
